/*
$Header: /vol/baal-vol3/projekt98/quellen/XCTL_32/DETECUSE/Detec_hw.cpp,v 1.17 2004/06/17 19:33:03 reinecke Exp $

Projekt   : XCTL
Subsystem : Detektoren
Autor     : Jan Picard <picard@informatik.hu-berlin.de> 2001-2002
          Institut fuer Informatik,
          Humboldt-Universitaet Berlin
Inhalt    : Code fuer die Hardwareklassen, die fuer die Messadapter
          die Kommunikation mit der Detektoren-Hardware erledigen
*/
#include "internls\evrythng.h"
#include "detecuse\detectorcontroller.h"

#include "utils\u_utils.h"
#include <fstream.h>
//ap #define STRICT


//--||--\\--||--//--||--\\--||--//--||--\\--||--//--||--\\--||--//--||--\\--||--

//Verwaltungsstrukturen (Treiber)
extern DeviceList DetectorDrivers;
extern ControllerList DetectorControllers;

//###########################################################################
// TRadiconController
//###########################################################################

// storage specifiations
const int LNGMSG= 25;     // length of buffer receive/transmit message
const int LNGPRG= 0x7f0;  // length of buffer program

// detector internals
// Timeout: Der urspruenglich verwendete Timeout-Wert war 3 'ticks',
// 1 tick= 1/18.2 sec= ca. 55 ms
const int TMOUT= 3*55;

const long KMGHZ= 500000L;   // frequency base (affects exposure time)
const int NCYCL= 8;         // number of retries for messaging

//  unknown codes
const unsigned char MONIT= 0x93;      // monitor controller???
const unsigned char PROGR= 0x6c;      // program ready???

enum responseCode
{
	ACK= 0x06,     // acknowledged
	NAK= 0x15,       // not acknowledged
	CAN= 0x18,       // canceled (not acknowledged)
	DLE= 0x10,       // data link escape
	EOT= 0x04,       // end of transmission
};

//! controller function codes
enum EFunctionCode {
	CF0= 0x00,     // set parameters
	CF1= 0x01,       // count pulses during exposure time
	CF2= 0x02,       // count time when number of pulses is set
	CF3= 0x03,       // intensimeter
	CF5= 0x05,       // stop controller
	CF9= 0x09,       // receive result time and pulses
	CFB= 0x0b,       // receive current time and pulses
	CFF= 0x0f     // load program to card ???
};

/////////////////////////////////////////////////////////////////
TRadiconController::TRadiconController(EDeviceType _DeviceID, LPTSTR _HardwareID, DeviceList* Devices) 
						: Controller(_DeviceID,_HardwareID,1), dataPort(0), controlPort(1)
{
	char* szDeviceName= Devices->GetFileName(DeviceID, HardwareID);
	if(!szDeviceName)
	{
		Hardware= new DummyIo(DeviceID,HardwareID);
	}
	else
	{
		Hardware= new HardwareIo(szDeviceName);
	}
}

TRadiconController::~TRadiconController()
{
	_FREEOBJ(Hardware);
}

BOOL TRadiconController::Check()
{
	//ap - noch zu implementieren!!!
	return(TRUE);
}

/*******************************************************************************
  procedure: SetParameters
  purpose:   upload parameters to controller
  result:    0= operation successful
             -1= communication error
             -2= controller can't make this function to present time or error
                  code function
             -3= parameter error
******************************************************************************/
int TRadiconController::SetParameters( unsigned short upperThreshold,
                                       unsigned short lowerThreshold,
                                       int highvoltage, double exposureTime,
                                       unsigned long impulseCount, BOOL bSound )
{
	unsigned char bufmsg[LNGMSG];   // buffer receive/transmit message
	memset(bufmsg, 0x0, LNGMSG);

	// format energy integral
	bufmsg[0]= 0x1;

	// format sound
	bufmsg[1]= (unsigned char) (bSound ? 7 : 8);

	// format high voltage do bitshift to get next Byte
	bufmsg[7]= (unsigned char) highvoltage;
	highvoltage= (highvoltage >> 8) & 0xff;
	bufmsg[8]= (unsigned char) highvoltage;

	// format lower threshold do bitshift to get next Byte
	bufmsg[9]= (unsigned char) lowerThreshold;
	lowerThreshold= (lowerThreshold >> 8) & 0xff;
	bufmsg[10]= (unsigned char) lowerThreshold;

	// format upper threshold do bitshift to get next Byte
	bufmsg[11]= (unsigned char) upperThreshold;
	upperThreshold= (upperThreshold >> 8) & 0xff;
	bufmsg[12]= (unsigned char) upperThreshold;

	// determine and format total counts
	// do bitshifts to get next bytes
	unsigned long wtmr= (unsigned long)(exposureTime * KMGHZ);      // !! Typkonvertierung !!
	bufmsg[13]= (unsigned char)(wtmr & 0xff);
	bufmsg[14]= (unsigned char)((wtmr >> 8) & 0xff);
	bufmsg[15]= (unsigned char)(wtmr >> 16) & 0xff;
	bufmsg[16]= (unsigned char)(wtmr >> 24) & 0xff;

	// format impulse count
	// do bitshifts to get next bytes
	bufmsg[17]= (unsigned char)(impulseCount & 0xff);
	bufmsg[18]= (unsigned char)((impulseCount >> 8) & 0xff);
	bufmsg[19]= (unsigned char)(impulseCount >> 16) & 0xff;
	bufmsg[20]= (unsigned char)(impulseCount >> 24) & 0xff;

	// format interrupt
	bufmsg[21]= 0;

	// transmit message for NCYCL cycles or until return value is 0
	int j;

	for (int i= 0; i < NCYCL; i++ )
	{
		j= TransmitMessage( CF0, bufmsg, LNGMSG );

		if (j == 0)
			break;

		if (j == -2)
			return -2;
	}

	// if max number of cycles reached and return value is not 0
	if (j != 0)
		return -1;

	return 0;
}

/*******************************************************************************
  procedure: Execute
  purpose:   issue command to controller
  result:     0= operation successful
             -1= communication error
             -2= controller can't make this function to present time or error
                  code function
             -3= parameter error (unknown command)
*******************************************************************************/
int TRadiconController::Execute( EOperationMode fun )
{
	register int i, j;
	EFunctionCode eFunctionCode;
	
	switch (fun) 
	{ // depending on the command to issue... (function code filter)		
		case fixedExposureCounts: // count impulses during exposure time
			eFunctionCode= CF1; 
			break;

		case fixedImpulseCounts: // count time when number of impulses is set
			eFunctionCode= CF2; 
			break;

		case intensimeter: // intensimeter
			eFunctionCode= CF3; 
			break;

		case stop: // stop controller
			eFunctionCode= CF5; 
			break;

		default: // unknown command
			return -3;
	}

	// transmit message NCYCL cycles or until return value is 0
	for ( i= 0; i < NCYCL; i++ )
	{
		j= TransmitMessage( eFunctionCode);

		if (j == 0)
		  break;

		if (j == -2)
		  return -2;
	}

	if (j != 0)
		return -1;   // unknown return code

	return 0;
}

/*******************************************************************************
  procedure: GetValues
  purpose:   get information from controller
  result:    0= operation successful
             -1= communication error
             -2= controller can't make this function to present time or error
                  code function
             -3= parameter error
             -4= data isn't ready
******************************************************************************/
int TRadiconController::GetValues( EReadoutMode fun, double *ftmr, unsigned long *imp )
{
	register int i, j;
	double wftmr;
	unsigned long wimp, wtmr, ww;
	unsigned char * p;
	int l;
	/* buffer receive message */
	unsigned char bufmsg[LNGMSG];

	EFunctionCode eFunctionCode;

	switch (fun)
	{
		case final: // receive result time and impulses
			eFunctionCode= CF9; 
			break;

		case intermediate: // receive current time and impulses
			eFunctionCode= CFB; 
			break;

		default: // parameter error
			return -3;
	}

	for ( i= 0; i < NCYCL; i++ )
	{
		j= ReceiveMessage( eFunctionCode, bufmsg, LNGMSG, l );

		if ( j == 0 )
			break;

		if ( (j == -2) || (j == -4) )
			return j;
	}

	if ( j != 0 )
		return -1;

	// read number of counts do Bitshifts to cycle through
	// the four Bytes of the bufmsg that represent the longint
	p= bufmsg;

	wtmr= 0;
	ww= 0;
	ww= *p;
	wtmr= wtmr + ww;
	p++;

	ww= 0;
	ww= *p;
	ww= ww << 8;
	wtmr= wtmr + ww;
	p++;

	ww= 0;
	ww= *p;
	ww= ww << 16;
	wtmr= wtmr + ww;
	p++;

	ww= 0;
	ww= *p;
	ww= ww << 24;
	wtmr= wtmr + ww;
	p++;

	// turn number of counts into exposure time do Bitshifts to cycle
	// through the four Bytes of the bufmsg that represent the longint
	wftmr= wtmr;
	*ftmr= wftmr / (KMGHZ);

	// read impulse count
	wimp= 0;

	ww= 0;
	ww= *p;
	wimp= wimp + ww;
	p++;

	ww= 0;
	ww= *p;
	ww= ww << 8;
	wimp= wimp + ww;
	p++;

	ww= 0;
	ww= *p;
	ww= ww << 16;
	wimp= wimp + ww;
	p++;

	ww= 0;
	ww= *p;
	ww= ww << 24;
	wimp= wimp + ww;
	*imp= wimp;

	return 0;
}

/*******************************************************************************
  procedure: reset
  purpose:   reset controller
  synopsis:  void CALLTYPE reset( )
  result:    nothing
*******************************************************************************/
void TRadiconController::reset()
{
	/* first write state register (port address, that identifies controller) into
	 data register. then write lower byte of accumulator register to the data
	 register
	_asm {
		mov dx, rs;
		out dx, al;
	}
	*/

	// JP!! Achtung, es ist voellig unklar, welcher Wert geschrieben werden soll,
	// da AL nicht initialisiert wurde, schreibe ich hier (willkuerlich) eine 0
	Write(controlPort,BYTE(0));
}

/******************************************************************************
 *  Funktionscode zum Controller uebertragen                                  *
 *                                                                            *
 *  Rueckgabewert:  0   - OK                                                  *
 *                 -1   - FEHLER                                              *
 *                 -2   - Controller kann die angeforderte Funktion zu diesem *
 *                        Zeitpunkt nicht ausfuehren                          *
 *                        oder unbekannter Funktionscode                      *
 *                 -3   - Controller zum FirmwareUpload bereit                *
 *                 -4   - Controller hat Kommunikation abgebrochen            *
 ******************************************************************************/
int TRadiconController::TransmitFunctionCode(int nFunctionCode)
{
	//Datenregister einlesen...zu welchem Zweck ???
	//eingelesener Wert wird nicht weiter verwendet...
	//dataPort.In8Bit();	ap
	BYTE data;
	Read(dataPort,&data);

	//Funktionscode zum Controller uebertragen

	if (out_byte(nFunctionCode) != TRUE)
		return -1;

	//Antwort des Controllers empfangen
	unsigned char chControllerAnswer;

	if (in_byte(chControllerAnswer) != TRUE)
		return -1;

	//Controllerantwort != ACK => Fehler spezifizieren...
	if (chControllerAnswer != ACK)
	{
		//???
		if (nFunctionCode == CFF && chControllerAnswer == PROGR)
			return -3;

		//Controllerantwort= CAN (Kommunikation abbrechen)
		if (chControllerAnswer == CAN)
			return -4;

		//Controllerantwort= NAK (unbekannter Funktionscode)
		if (chControllerAnswer == NAK)
			return -2;

		//unspezifische Controllerantwort
		return -1;
	}

	return 0;
}

/******************************************************************************
 *  Nachricht (Funktionscode und Daten) zum Controller uebertragen            *
 *                                                                            *
 *  Parameter:                                                                *
 *    int nFunctionCode           - Funktionscode (Nachricht)                 *
 *    unsigned char* lpszMessage  - Zeiger auf Datenpuffer                    *
 *    int nMessageLength          - Anzahl der zu uebertragenden Bytes        *
 *                                                                            *
 *  Rueckgabewert:  0   - OK                                                  *
 *                 -1   - FEHLER                                              *
 *                 -2   - Controller kann die angeforderte Funktion zu diesem *
 *                        Zeitpunkt nicht ausfuehren                          *
 *                        oder unbekannter Funktionscode                      *
 *                 -3   - Checksummenfehler                                   *
 ******************************************************************************/
int TRadiconController::TransmitMessage(int nFunctionCode, unsigned char* lpszMessage,
                                int nMessageLength)
{
  int nRetCode= TransmitFunctionCode(nFunctionCode);

  if ( nRetCode != 0)
  {
    if ( nRetCode == -3)
      return 0;

    return nRetCode;
  }

  int nChecksum= 0;  //JP!! Test: Checksumme als UCHAR

  //Datenpuffer byteweise zum Controller uebertragen...

  for (int i= 0; i < nMessageLength; i++)
  { //Byte aus dem Datenpuffer zum Controller uebertragen

    if (out_byte(*lpszMessage) != TRUE)
      return -1;

    //gesendete Bytes aufsummieren (fuer Checksumme)
    nChecksum += *lpszMessage;

    //Wenn ein zu uebertragendes Byte den gleichen Wert wie der Steuercode DLE
    //besitzt, so mu dieses Byte zweimal hintereinander uebertragen werden,
    //um die Steuercodebedeutung aufzuheben
    if (*lpszMessage == 0x10) //*p == DLE
    { //zweites senden des Bytes

      if (out_byte(*lpszMessage) != TRUE)
        return -1;

      //gesendete Bytes aufsummieren (fuer Checksumme)
      nChecksum += *lpszMessage;
    }

    //Pufferzeiger inkrementieren
    lpszMessage++;
  }

  //Nachricht DLE zum Controller uebertragen => Daten vollstaendig uebertragen
  if (out_byte(DLE) != TRUE)
    return -1;

  //gesendet Bytes aufsummieren (fuer Checksumme)
  nChecksum += DLE;

  //Nachricht EOT zum Controller uebertragen
  //=> Nachrichtenuebertragung abschliessen
  if (out_byte(EOT) != TRUE)
    return -1;

  nChecksum += EOT; //gesendete Bytes aufsummieren (fuer Checksumme)
  nChecksum= ~nChecksum;   //bitweise Negation (1er-Komplement)
  nChecksum++;      //=> 2er-Komplement

  //ermittelte Checksumme zum Controller uebertragen
  if (out_byte(nChecksum) != TRUE)
    return -1;

  unsigned char chControllerAnswer;

  if (in_byte(chControllerAnswer) != TRUE)
    return -1;

  if (chControllerAnswer == NAK)
    return -3;      // NAK= Checksummenfehler

  return 0; //Funktion mit OK beenden
}

/******************************************************************************
 *  Nachricht vom Controller empfangen                                        *
 *  (Funktionscode dem Controller uebermitteln und Ergebnisdaten empfangen)   *
 *                                                                            *
 *  Parameter:                                                                *
 *    int nFunctionCode           - Funktionscode (Nachricht)                 *
 *    unsigned char* lpszMessage  - Zeiger auf Datenpuffer                    *
 *    int nMaxMessageLength       - max. Laenge der zu empfangenden Daten     *
 *    int& nRealMessageLength     - Zeiger auf Anzahl der empfangenen Daten   *
 *                                                                            *
 *  Rueckgabewert:  0   - OK                                                  *
 *                 -1   - FEHLER                                              *
 *                 -2   - Controller kann die angeforderte Funktion zu diesem *
 *                        Zeitpunkt nicht ausfuehren                          *
 *                        oder unbekannter Funktionscode                      *
 *                 -3   - Checksummenfehler                                   *
 *                 -4   - Daten stehen nicht bereit                           *
 ******************************************************************************/
int TRadiconController::ReceiveMessage(int nFunctionCode, unsigned char* lpszMessage,
                               int nMaxMessageLength, int& nRealMessageLength)
{
  int nRetCode= TransmitFunctionCode(nFunctionCode);

  if ( nRetCode != 0)
  {
    if ( nRetCode == -3)
      return 0;

    return nRetCode;
  }

  int nChecksum= 0;
  nRealMessageLength= 0; //Anzahl der empfangenen Daten mit 0 initialisieren

  //Daten byteweise vom Controller empfangen bis die max. Anzahl erreicht ist
  //oder die Datenuebertragung durch DLE vom Controller beendet wird
  unsigned char chControllerAnswer1, chControllerAnswer2;

  while (nRealMessageLength < nMaxMessageLength)
  { //Byte vom Controller empfangen

    if (in_byte(chControllerAnswer1) != TRUE)
      return -1;

    //empfangene Bytes aufsummieren (fuer Checksumme)
    nChecksum += chControllerAnswer1;

    //empfangenes Byte= DLE => testen ob Steuercode(DLE) oder Datenwert(10h)
    if (chControllerAnswer1 == DLE)
    {  //Folgebyte vom Controller empfangen

      if (in_byte(chControllerAnswer2) != TRUE)
        return -1;

      nChecksum += chControllerAnswer2;

      //empfangenes Byte= EOT (Ende der Uebertragung) => vorheriges Byte
      //war Steuercode DLE => Datenempfang vorzeitig abbrechen
      if (chControllerAnswer2 == EOT)
        break;

      //empfangenes Byte != DLE => vorheriges Byte war weder Steuercode(DLE)
      //noch Datenwert(10h) => Funktion mit FEHLER beenden
      if (chControllerAnswer2 != DLE)
        return -1;
    }

    //empfangenes Byte in den Datenpuffer uebernehmen
    *lpszMessage= chControllerAnswer1;

    //Pufferzeiger inkrementieren
    lpszMessage++;

    //Anzahl der empfangenen Daten inkrementieren
    nRealMessageLength++;
  }

  //letztes Byte vom Controller empfangen (zur Ueberpruefung der Checksumme)
  if (in_byte(chControllerAnswer1) != TRUE)
    return -1;

  //Checksumme pruefen
  nChecksum= (nChecksum + chControllerAnswer1) & 0xff;

  //Checksumme != 0 => NAK (wegen Checksummenfehler) an den Controller senden
  if (nChecksum != 0)
  { //Nachricht NAK (Checksummenfehler) an den Controller uebermitteln

    if (out_byte(NAK) != TRUE)
      return -1;

    //Funktion mit Fehlercode -3 beenden
    return -3;
  }

  //Nachricht ACK (Checksumme OK) an Controller uebermitteln
  if (out_byte(ACK) != TRUE)
    return -1;

  return 0;
}

/******************************************************************************
 *  Datenbyte zum Controller uebertragen                                      *
 ******************************************************************************/
BOOL TRadiconController::out_byte(unsigned char d)
{
  unsigned long dwStart, dwNow;

  //(1) aktuelle Systemzeit ermitteln
  dwStart= dwNow= CurrentTime();

  //Warten auf 'Bereit-Zustand' des Controllers innerhalb einer bestimmten Zeit
  //(TIMEOUT-Schleife)

  while (( dwNow - dwStart ) < TMOUT)
  {
    //Controllerstatus ermitteln

    if (IsReadyToWrite() == TRUE)
    {
      //Byte vom Controller empfangen (Datenregister lesen)
      //dataPort.Out8Bit(d);	ap
		Write(dataPort,(BYTE)d);
      return TRUE;
    }

    //(2) aktuelle Systemzeit ermitteln
    dwNow= CurrentTime();

    //Fehler: zuletzt gemessene Systemzeit(2) liegt vor der ersten gemessenen
    //Systemzeit(1) => Anpassen der ersten Systemzeit(1)
    //=> bisher verstrichene Zeit wird zurueckgesetzt => TIMEOUT beginnt erneut
    if (dwNow < dwStart)
      dwStart= dwNow;
  }

  //Daten stehen nicht bereit
  return FALSE;
}

/******************************************************************************
 *  Datenbyte vom Controller empfangen                                        *
 ******************************************************************************/
BOOL TRadiconController::in_byte(unsigned char &d)
{
  unsigned long dwStart, dwNow;

  //(1) aktuelle Systemzeit ermitteln
  dwStart= dwNow= CurrentTime();

  //Warten auf 'Bereit-Zustand' des Controllers innerhalb einer bestimmten Zeit
  //(TIMEOUT-Schleife)

  while (( dwNow - dwStart ) < TMOUT)
  {
    //Controllerstatus ermitteln

    if (IsReadyToRead() == TRUE)
    {
      //Byte vom Controller empfangen (Datenregister lesen)
      //d= dataPort.In8Bit();	ap
		Read(dataPort,&d);
      return TRUE;
    }

    //(2) aktuelle Systemzeit ermitteln
    dwNow= CurrentTime();

    //Fehler: zuletzt gemessene Systemzeit(2) liegt vor der ersten gemessenen
    //Systemzeit(1) => Anpassen der ersten Systemzeit(1)
    //=> bisher verstrichene Zeit wird zurueckgesetzt => TIMEOUT beginnt erneut
    if (dwNow < dwStart)
      dwStart= dwNow;
  }

  //Daten stehen nicht bereit
  return FALSE;
}

/******************************************************************************
 *  Testen ob Controller zur Datenuebergabe (senden) bereit                   *
 *                                                                            *
 *  Rueckgabewert:  TRUE   - Controller bereit (Datenregister nicht leer)     *
 *                  FALSE  - Controller nicht bereit (Datenregister leer)     *
 ******************************************************************************/
BOOL TRadiconController::IsReadyToRead()
{ 
  //Steuerregister lesen und auswerten...
  //JP Achtung: Der Controller ist zum Auslesen bereit,
  //JP wenn das Ergebnis != 0 ist. Beim Bereitschaftstest zum Senden ist der
  //JP Controller bereit, wenn das Ergebnis == 0 ist
  //return (( controlPort.In8Bit() & 0x02 ) != 0);	ap
	BYTE data;
	Read(controlPort,&data);
	return(( data & 0x02 ) != 0);
}

/******************************************************************************
 *  Testen ob Controller zur Datenuebernahme (empfangen) bereit               *
 *                                                                            *
 *  Rueckgabewert:  TRUE   - Controller bereit (Datenregister nicht leer)     *
 *                  FALSE  - Controller nicht bereit (Datenregister leer)     *
 ******************************************************************************/
BOOL TRadiconController::IsReadyToWrite()
{ 
  //Steuerregister lesen und Status auswerten...
  //return (( controlPort.In8Bit() & 0x01 ) == 0);	ap
	BYTE data;
	Read(controlPort,&data);
	//rh return(( data & 0x01 ) != 0);
	return(( data & 0x01 ) == 0);

}

/*******************************************************************************
--- UploadFirmware ---
upload controller firmware
 
return values:   0= operation successful
                -1= communication error
                -2= controller rejects connect
                -3= error using firmware file
                -4= unspecified error
*******************************************************************************/
int TRadiconController::UploadFirmware(void)
{
  // open firmware file
  char szFileName[MaxString];
  sprintf(szFileName, "%sscs.prg", GetDirectory());
  ifstream ifile(szFileName, ios::in | ios::binary);

  if ( ifile.fail() )
  {
    MessageBox(0, "Kann Radicon-Firmwaredatei scs.prg nicht oeffnen!",
               "Error", MBINFO );
    return -3;
  }

  // check file size
  ifile.seekg (0, ios::end);

  int nFileSize= ifile.tellg();

  ifile.seekg (0, ios::beg);

  if ( nFileSize != LNGPRG )
  {
    MessageBox(0, "Die Radicon-Firmwaredatei scs.prg hat nicht die korrekte Laenge!",
               "Error", MBINFO );
    return -3;
  }

  unsigned char* pBuf= new unsigned char[LNGPRG];

  if ( pBuf == NULL )
  {
    MessageBox(0, "Nicht genug Speicher!",
               "Error", MBINFO );
    return -4;
  }

  // read content of file into memory
  ifile.read(pBuf, LNGPRG);

  ifile.close();

  // This method reads bytes from controller multiple times, but can't
  // say why. This behavior is preserved from the old C-sources, because
  // I don't know the Radicon communication protocol.

  unsigned char d;

  in_byte(d);    // read a byte from controller - why???

  Execute(stop);

  // send firmware to controller
  int nRetCode;

  int i;

  for ( i= 0; i < NCYCL; i++ )
  {
    nRetCode= TransmitMessage( CFF, pBuf, LNGPRG );

    if (nRetCode == -2)
    {
      _FREEOBJ(pBuf);
      return -2;
    }

    if (nRetCode == 0)
      break;
  }

  _FREEOBJ(pBuf);

  if ( i == NCYCL )
    return -1;    // Timeout erreicht

  if ( d == MONIT )
    in_byte(d);

  return 0;
}

/******************************************************************************
 *  Zeit seit dem Start des Systems in Millisekunden ermitteln                *
 *                                                                            *
 *  Rueckgabewert:  dword - Systemzeit in Millisekunden                       *
 ******************************************************************************/
DWORD TRadiconController::CurrentTime()
{
  return GetTickCount();
}


//###########################################################################
// TAm9513
//###########################################################################

const int IoWaiTAm9513= 43;
// Chip
const BYTE CHIP1= 0x0;
const BYTE CHIP2= 0x1;
const BYTE DIGITAL= 0x2;

const BYTE BIT= 0x4;
const BYTE CODE= 0x5;

// IO-Addressen Offset
const int DATA= 0x0;
const int CMD= 0x1;

// Gruppen
const BYTE CNT1= 0x01;
const BYTE CNT2= 0x02;
const BYTE CNT3= 0x03;
const BYTE CNT4= 0x04;
const BYTE CNT5= 0x05;
const BYTE CTRG= 0x07;

// Zaehler-Id Bit's
const BYTE CTB1= 0x01;
const BYTE CTB2= 0x02;
const BYTE CTB3= 0x04;
const BYTE CTB4= 0x08;
const BYTE CTB5= 0x10;

// Register der Gruppen
const BYTE CMR= 0x00; // Mode-Register
const BYTE CLR= 0x08; // Load-Register
const BYTE CHR= 0x10; // Hold-Register
const BYTE CHIR= 0x18; // Hold+Inc-Register
const BYTE ALM1= 0x00; // Alarm-Register 1
const BYTE ALM2= 0x08; // Alarm-Register 2
const BYTE MMR= 0x10; // Master-Mode-Register
const BYTE STS= 0x18; // Status-Register (8 Bit)
// Funktions-Bit fuer CMR
const WORD CB_UP= 3; // aufwaerts zaehlen
const WORD CB_BCD= 4; // BCD zaehlen
const WORD CB_REPEAT= 5; // automatisch neu starten
const WORD CB_RLorH= 6; // laden von L-Reg oder H-Reg
const WORD CB_ENGATE= 7; // special gate
const WORD CB_FALLEG= 12; // mit steigender Flanke zaehlen
// Funktions-Bit fuer MMR
const WORD MB_CMP1E= 2;
const WORD MB_CMP2E= 3;
const WORD MB_NFOUT= 12;
const WORD MB_16BIT= 13;
const WORD MB_DIS_INC= 14;
const WORD MB_BCD_DIV= 15;
// Funktions-Code fuer CMR (nur Setzen mit |)
const WORD CC_CNTM1= 0x0000;
const WORD CC_NOGATE= 0x0000;
const WORD CC_AHTCNM1= 0x2000;
const WORD CC_AHNP1= 0x4000;
const WORD CC_AHNM1= 0x6000;
const WORD CC_AHN= 0x8000;
const WORD CC_ALN= 0xA000;
const WORD CC_AHEGN= 0xC000;
const WORD CC_ALEGN= 0xE000;
// Output-Control
const WORD CC_I_OLow= 0x0000;
const WORD CC_A_HTCP= 0x0001;
const WORD CC_TCTOG= 0x0002;
const WORD CC_I_OTRI= 0x0004;
const WORD CC_A_LTCP= 0x0005;

// Quellen-Bezeichner fuer CMR und MMR
const WORD C_CNM1= 0x0; // F1 MMR
const WORD C_SRC1= 0x1;
const WORD C_SRC2= 0x2;
const WORD C_SRC3= 0x3;
const WORD C_SRC4= 0x4;
const WORD C_SRC5= 0x5;
const WORD C_GATE1= 0x6;
const WORD C_GATE2= 0x7;
const WORD C_GATE3= 0x8;
const WORD C_GATE4= 0x9;
const WORD C_GATE5= 0xA;
const WORD C_F1= 0xB;
const WORD C_F2= 0xC;
const WORD C_F3= 0xD;
const WORD C_F4= 0xE;
const WORD C_F5= 0xF;


//******** Methoden fr den Zhler-IC Am9513A **********************************
//Wieviele Detektoren knnen an diesen Controller angeschlossen werden???
TGenericController::TGenericController(EDeviceType _DeviceID, LPTSTR _HardwareID, DeviceList* Devices) 
						: Controller(_DeviceID,_HardwareID,5),bIsStopped(FALSE),bIsLatched(FALSE),
						  wLowTicks(1000),wHighTicks(100),wLowCounts(1000),wHighCounts(100),
						  nbChipId(0)
{
	char* szDeviceName= Devices->GetFileName(DeviceID, HardwareID);
	if(!szDeviceName)
	{
		Hardware= new DummyIo(DeviceID,HardwareID);
	}
	else
	{
		Hardware= new HardwareIo(szDeviceName);
	}
}

TGenericController::~TGenericController()
{
	_FREEOBJ(Hardware)
}

BOOL TGenericController::Check()
{
	//ap - noch zu implementieren!!!
	return(TRUE);
}

int TGenericController::Init(float _fTimeCorrection)
{
	//ap
	fTimeCorrection= _fTimeCorrection;

	// Kommunikations-Test
	ChooseDataPtr( CNT1, CMR );
	WriteData( 0xA121 );
	ChooseDataPtr( CNT1, CMR );

	if ( ReadData() != 0xA121 )
	return FALSE;

	// Zhler zurcksetzen
	Reset();
	WriteCmd( 0xF9 );

	// Konfigurieren des Mastermode-Register
	ChooseDataPtr( CTRG, MMR );
	WriteData( 0xCA00 ); // F1 --> FOUT

	// Konfigurieren des TimeTick-Zhler
	ChooseDataPtr( CNT1, CMR );
	WriteData( 0xA121 );

	ChooseDataPtr( CNT2, CMR );
	WriteData( 0x0202 );

	// Konfigurieren des Impuls-Zhler
	ChooseDataPtr( CNT3, CMR );
	WriteData( 0xA321 );

	ChooseDataPtr( CNT4, CMR );
	WriteData( 0x0402 );

	// Konfigurieren des Akustik-Zhler
	ChooseDataPtr( CNT5, CMR );
	WriteData( 0x0C21 );

	return TRUE;
}

int TGenericController::IOCTL( EIOCCmd cmd )
{
	DWORD dwData= 0;
	return IOCTL( cmd, dwData );
}

int TGenericController::IOCTL( EIOCCmd cmd, DWORD& param )
{
	WORD regH, regL;
	WORD wValue;

	switch ( cmd )
	{
		case iocSetTimeTicks:
			// Zerlegung in ein Produkt;
			param= ( DWORD ) ( ( double ) param * fTimeCorrection );
			SplitNumber( param, wLowTicks, wHighTicks );
			ChooseDataPtr( CNT1, CLR );
			WriteData( wLowTicks );
			ChooseDataPtr( CNT2, CLR );
			WriteData( wHighTicks ); // TC Toggle
			param= ( DWORD ) wLowTicks * wHighTicks;
			param= ( DWORD ) ( ( double ) param / fTimeCorrection );
			break;

		case iocSetCounts:
			SplitNumber( param, wLowCounts, wHighCounts );
			param= ( DWORD ) wLowCounts * wHighCounts;
			ChooseDataPtr( CNT3, CLR );
			WriteData( wLowCounts );
			ChooseDataPtr( CNT4, CLR );
			WriteData( wHighCounts ); // TC Toggle
			break;

		case iocStopCCDTiming:
			DisarmC( CTB1 | CTB2 );
			SetToggleOut( CNT2 );
			bIsStopped= TRUE;
			break;

		case iocStartCCDTiming:
			if ( !bIsStopped )
				DisarmC( CTB1 | CTB2 );

			ChooseDataPtr( CNT1, CMR );
			// wValue= 0xA121;
			wValue= 0x0121;
			WriteData( wValue );

			ChooseDataPtr( CNT2, CMR );
			// wValue= 0x0202;
			wValue= 0x0222;
			WriteData( wValue );

			bIsStopped= FALSE;
			bIsLatched= FALSE;

			LoadC( CTB1 | CTB2 );
			LoadC( CTB1 | CTB2 );

			SetToggleOut( CNT2 );
			SetToggleOut( CNT4 );

			ClearToggleOut( CNT1 );
			ClearToggleOut( CNT3 );

			ArmC( CTB1 | CTB2 );
			break;

		case iocStartCCDShot:
			if ( !bIsStopped )
				DisarmC( CTB1 | CTB2 );

			bIsStopped= FALSE;
			bIsLatched= FALSE;

			LoadC( CTB1 | CTB2 );
			LoadC( CTB1 | CTB2 );

			SetToggleOut( CNT2 );
			SetToggleOut( CNT4 );

			ClearToggleOut( CNT1 );
			ClearToggleOut( CNT3 );

			ArmC( CTB1 | CTB2 );
			break;

		case iocReadTicks:
			if ( !bIsLatched )
			{
				DisarmC( CTB1 | CTB2 | CTB3 | CTB4 );
				bIsStopped= TRUE;
				LatchToHoldC( CTB1 | CTB2 | CTB3 | CTB4 );
				bIsLatched= TRUE;
			}
			ChooseDataPtr( CNT1, CHR );
			regL= ReadData( );
			ChooseDataPtr( CNT2, CHR );
			regH= ReadData( );
			param= ( DWORD ) ( wHighTicks - regH ) * ( DWORD ) wLowTicks;
			param += ( DWORD ) ( wLowTicks - regL );
			param= ( DWORD ) ( ( double ) param / fTimeCorrection );
			break;

		case iocReadCounts:
			if ( !bIsLatched )
			{
				DisarmC( CTB1 | CTB2 | CTB3 | CTB4 );
				bIsStopped= TRUE;
				LatchToHoldC( CTB1 | CTB2 | CTB3 | CTB4 );
				bIsLatched= TRUE;
			}

			ChooseDataPtr( CNT3, CHR );
			regL= ReadData( );
			ChooseDataPtr( CNT4, CHR );
			regH= ReadData( );
			param= ( DWORD ) ( wHighCounts - regH ) * ( DWORD ) wLowCounts;
			param += ( DWORD ) ( wLowCounts - regL );
			break;

		case iocConfigureAcustic:
			ChooseDataPtr( CNT5, CLR );
			WriteData( ( WORD ) param );
			LoadAndArmC( CTB5 );
			break;

		case iocStartCounting:
			DisarmC( CTB1 | CTB2 | CTB3 | CTB4 );
			bIsStopped= FALSE;
			bIsLatched= FALSE;
			LoadC( CTB1 | CTB2 | CTB3 | CTB4 );
			DelayTime( 1 );
			LoadC( CTB1 | CTB2 | CTB3 | CTB4 );
			ClearToggleOut( CNT1 );
			ClearToggleOut( CNT3 );
			SetToggleOut( CNT2 );
			SetToggleOut( CNT4 );
			DelayTime( 1 );
			ArmC( CTB1 | CTB2 | CTB3 | CTB4 );
			break;

		case iocIsCountingReady:
			switch ( 0x14 & ReadStatus( ) )
			{
				case 0x14: 
					param= 0; 
					break; // no counter ready

				case 0x10: 
					param= 1; 
					break; // time counter ready

				case 0x04: 
					param= 2; 
					break; // impulse counter ready

				default: 
					param= 3;
			}
			return (int)param;

		case iocStopCounting:
			DisarmC( CTB1 | CTB2 | CTB3 | CTB4 );
			bIsStopped= TRUE;
			break;
	}

	return TRUE;
}

BOOL TGenericController::SplitNumber( DWORD value, WORD& low, WORD& high )
{
	// in case of any changes notify the function IOCTL()
	if ( value < 0xF000 )
	{
		high= 1;
		low= ( WORD ) value;
		return TRUE;
	}

	long f= value;
	long f1= 4;
	long f2= value / f1;
	int cnt= 1000;
	long fail;

	while ( cnt-- )
	{
		fail= f - f1 * f2;

		if ( ( f1 > 0xE000 ) || ( f2 > 0xE000 ) )
		{
			if ( f1 < f2 )
			{
				f1 *= 2;
				f2 /= 2;
			}
			else
			{
				f2 *= 2;
				f1 /= 2;
			}
			continue;
		}
		if ( fail == 0 )
			break;
		
		if ( f1 < f2 )
			f2 += fail / f1;
		else
			f1 += fail / f2;

		if ( ( f1 > fail ) && ( f2 > fail ) )
			break;
	} // end-while

	if ( f1 < f2 )
	{
		// f1 und f2 vertauschen
		long swap= f1;
		f1= f2;
		f2= swap;
	}

	if ( f1 < ( fail * 2 ) )
		f2++;
	
	high= ( WORD ) f1;
	low= ( WORD ) f2;
	
	if ( cnt == 0 )
		MessageBox( GetFocus( ), "Fehler beim Zerlegen !", "Am9513a", MBINFO );
	
	return TRUE;
}

DWORD TGenericController::GetTicksPerSecond(void)
{
	return 100000L;
}

void TGenericController::DigitalOut(BYTE data)
{
	//Portzugriffe nur ber Offsets zum Basisport im Treiber
	Hardware->Write(3,data);
}

void TGenericController::DigitalIn( BYTE *data )
{
	//Portzugriffe nur ber Offsets zum Basisport im Treiber
	Hardware->Read(2,data);
}

//ap - neu !!!
void TGenericController::SetSound(BOOL param)
{
	Hardware->Write(3, (BYTE)(param?1:0));
}

BOOL TGenericController::SelectChip( /* BYTE */)
{
	// Zhler zurcksetzen
	Reset();
	Delay( 50 );
	// Check IO
	ChooseDataPtr( CNT1, CLR );
	WriteData( 0xDAAB );
	ChooseDataPtr( CNT2, CLR );
	WriteData( 0x0000 );
	ChooseDataPtr( CNT1, CLR );
	WORD value= ReadData( );

	if ( 0xDAAB != value )
	{
		char buf[ MaxString ];
		sprintf( buf, "SelectChip() gab %X zurueck", value );
		MessageBox( GetFocus( ), buf, "Detector-Fehler Am9513a", MBINFO );
		return FALSE;
	}

	return TRUE;
}

void TGenericController::ArmC( BYTE counter )
{
	WriteCmd( counter | 0x20 );
}

void TGenericController::LoadC( BYTE counter )
{
	WriteCmd( counter | 0x40 );
}

void TGenericController::LoadAndArmC( BYTE counter )
{
	WriteCmd( counter | 0x60 );
}

void TGenericController::DisarmAndSaveC( BYTE counter )
{
	WriteCmd( counter | 0x80 );
}

void TGenericController::LatchToHoldC( BYTE counter )
{
	WriteCmd( counter | 0xA0 );
}

void TGenericController::DisarmC( BYTE counter )
{
	WriteCmd( counter | 0xC0 );
}

void TGenericController::ClearToggleOut( BYTE counter )
{
	WriteCmd( counter | 0xE0 );
}

void TGenericController::SetToggleOut( BYTE counter )
{
	WriteCmd( counter | 0xE8 );
}

void TGenericController::Reset( void )
{
	WriteCmd( 0xFF );
}

void TGenericController::ChooseDataPtr( BYTE group, BYTE reg )
{
	// Register adressieren
	WriteCmd( group | reg );
}

WORD TGenericController::ReadStatus( void )
{
	Delay( IoWaiTAm9513 );

	//Portzugriffe nur ber Offsets zum Basisport im Treiber
	BYTE data= 0;
	Hardware->Read(CMD,&data);
	return(data);
}

void TGenericController::WriteCmd( BYTE cmd )
{
	if ( nbChipId == DIGITAL )
		return;

	Delay( IoWaiTAm9513 );

	//Portzugriffe nur ber Offsets zum Basisport im Treiber
	Hardware->Write(CMD,cmd);
}

void TGenericController::WriteData( WORD daten )
{
	// Erst niederwertiges, dann hoeherwertiges Byte auf Port addr ausgeben

	//Portzugriffe nur ber Offsets zum Basisport im Treiber
	Delay(IoWaiTAm9513);
	Hardware->Write(DATA,(BYTE)(daten & 0x00FF));
	Delay(IoWaiTAm9513);
	Hardware->Write(DATA,(BYTE)(((WORD)daten >> 8) & 0x00FF));
	Delay(IoWaiTAm9513);
}

WORD TGenericController::ReadData( void )
{
	// Erst niederwertiges, dann hoeherwertiges Byte von Port addr lesen

	//Portzugriffe nur ber Offsets zum Basisport im Treiber
	BYTE tmpL=0,tmpH=0;
	WORD value= 0;
	
	Delay(IoWaiTAm9513);
	Hardware->Read(DATA,&tmpL);
	Delay(IoWaiTAm9513);
	Hardware->Read(DATA,&tmpH);
	value= tmpH;
	value= value << 8;
	value += tmpL;
	Delay( IoWaiTAm9513 );
	return value;
}

//##############################################################################
//******************* Methoden fuer den BraunPSD ******************************
//##############################################################################

TBraunPsdController::TBraunPsdController(EDeviceType _DeviceID, LPTSTR _HardwareID, DeviceList* Devices) 
						: Controller(_DeviceID,_HardwareID,1), SpecialHardware(0)
{
	char* szDeviceName= Devices->GetFileName(DeviceID, HardwareID);
	if (!szDeviceName)
	{
		SpecialHardware= new BraunPsdDummyIo();
		Hardware= SpecialHardware;
	}
	else
	{
		SpecialHardware= new BraunPsdIo(szDeviceName);
		Hardware= SpecialHardware;
	}
}

TBraunPsdController::~TBraunPsdController()
{
	_FREEOBJ(Hardware);
}

BOOL TBraunPsdController::Check()
{
	//ap - noch zu implementieren!!!
	return TRUE;
}

BOOL TBraunPsdController::SetTimeout(unsigned long Cycles)
{
	if (!SpecialHardware)
		return FALSE;

	return (SpecialHardware->SetTimeout(Cycles));
}

BOOL TBraunPsdController::GetData(IrpParamsGetData *Params) // 31.06.2004 MEMCORRECT (IrpParamsGetData *Params)
{
	if (!SpecialHardware)
		return FALSE;

	return (SpecialHardware->GetData(Params));
}